////////////////////////////////////////////////////////////////////////////////
//DEXi: Implements a Multi-Attribute Decision Making algorithm developed by
//dr. Marko Bohanec (http://www-ai.ijs.si/MarkoBohanec/dex.html)
//
//Copyright (C) 2004 Asobi d.o.o. (www.asobi.si)
//Authors: dr. Marko Bohanec, Dusan Omercevic, Andrej Kogovsek
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//Lesser General Public License for more details.
//
//You should have received a copy of the GNU Lesser General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
////////////////////////////////////////////////////////////////////////////////
package si.DEXi;

import org.w3c.dom.*;

import java.io.IOException;
import java.io.OutputStream;

import java.util.*;


/**
 * Attribute class
 * @author Andrej Kogov\u0161ek, Du\u0161an Omer\u010Devi\u0107
 * @version 1.0
 */
public class Attribute {
    private Vector attributes;
    private String description;
    private Boolean completeness;
    private Boolean explicitness;
    private String name;
    private Vector function = null;
    private Value value;
    private Scale scale = null;

    /**
     * Constructor with element parameter
     * @param element XML object parameter
     */
    public Attribute(final Element element) {
        NodeList nodes = element.getChildNodes();
        attributes = new Vector();

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);

            if (node.getNodeName().equals("NAME")) {
                if (getName() != null) {
                    throw new IllegalArgumentException(
                        "Attribute: Name already exists");
                }

                setName(node.getFirstChild().getNodeValue());
            } else if (node.getNodeName().equals("DESCRIPTION")) {
                if (getDescription() != null) {
                    throw new IllegalArgumentException(
                        "Attribute: Description already exists");
                }

                setDescription(node.getFirstChild().getNodeValue());
            } else if (node.getNodeName().equals("SCALE")) {
                if (scale != null) {
                    throw new IllegalArgumentException(
                        "Attribute: Scale already exists");
                }

                scale = new Scale((Element) node);
            } else if (node.getNodeName().equals("FUNCTION")) {
                if (function != null) {
                    throw new IllegalArgumentException(
                        "Attribute: Functions already exists");
                }

                processFunction(node);
            } else if (node.getNodeName().equals("ATTRIBUTE")) {
                Attribute attribute = new Attribute((Element) node);
                attributes.addElement(attribute);
            }
        }

        if (getName() == null) {
            throw new IllegalArgumentException("Attribute: Name is undefined");
        }

        checkExplicitness();
        checkCompleteness();
    }

    /**
     * Process function
     * @param node Node object parameter
     */
    private void processFunction(final Node node) {
        NodeList nodes = node.getChildNodes();
        String low = null;
        String high = null;
        String entered = null;
        function = new Vector();

        for (int x = 0; x < nodes.getLength(); x++) {
            Node funNode = nodes.item(x);

            if (funNode.getNodeName().equals("LOW")) {
                low = funNode.getFirstChild().getNodeValue();
            } else if (funNode.getNodeName().equals("HIGH")) {
                high = funNode.getFirstChild().getNodeValue();
            } else if (funNode.getNodeName().equals("ENTERED")) {
                entered = funNode.getFirstChild().getNodeValue();
            }
        }

        if (high == null) {
            high = low;
        }

        Integer[] iLow = parseFunction(low);
        Integer[] iHigh = parseFunction(high);
        Boolean[] bEntered = null;

        if (entered != null) {
            bEntered = parseEntered(entered);
        }

        if ((iHigh.length != iLow.length)
                || ((bEntered != null) && (bEntered.length != iLow.length))) {
            throw new IllegalArgumentException(
                "Attribute: high.lenght != low.lenght");
        }

        for (int l = 0; l < iLow.length; l++) {
            Rule rule = new Rule(iLow[l], iHigh[l],
                    new Boolean((bEntered == null)
                        || bEntered[l].booleanValue()));
            function.addElement(rule);
        }
    }

    /**
     * Parse function from String
     * @param value String object parameter
     * @return Array of Integer objects
     */
    private static Integer[] parseFunction(final String value) {
        Integer[] integer = null;
        char[] chars = value.toCharArray();
        integer = new Integer[chars.length];

        for (int i = 0; i < chars.length; i++) {
            integer[i] = new Integer(new String(chars, i, 1));
        }

        return integer;
    }

    /**
     * Parse entered from String
     * @param value String object parameter
     * @return Array of Boolean objects
     */
    private static Boolean[] parseEntered(final String value) {
        Boolean[] entered = null;
        char[] chars = value.toCharArray();
        entered = new Boolean[chars.length];

        for (int i = 0; i < chars.length; i++) {
            String strEntered = new String(chars, i, 1);

            if (strEntered.equals("+")) {
                entered[i] = new Boolean(true);
            } else {
                entered[i] = new Boolean(false);
            }
        }

        return entered;
    }

    /**
     * Check completeness
     *
     */
    protected void checkCompleteness() {
        boolean isComplet = true;
        int product = 1;

        for (int i = 0; i < attributes.size(); i++) {
            Attribute attribute = (Attribute) attributes.get(i);

            if (attribute.getSize().intValue() != 0) {
                product = product * attribute.getSize().intValue();
            }

            isComplet = isComplet && attribute.getCompleteness().booleanValue();
        }

        isComplet = isComplet && (this.getSize().intValue() > 0);
        isComplet = isComplet
            && ((function == null) || (product == function.size()));
        setCompleteness(new Boolean(isComplet));
    }

    /**
     * Check explicitness
     *
     */
    protected void checkExplicitness() {
        boolean isExplicit = true;

        for (int a = 0; a < attributes.size(); a++) {
            Attribute attribut = (Attribute) attributes.get(a);
            isExplicit = isExplicit
                && attribut.getExplicitness().booleanValue();
        }

        if (function != null) {
            for (int f = 0; f < function.size(); f++) {
                Rule rule = (Rule) function.get(f);
                isExplicit = isExplicit
                    && rule.getExplicitness().booleanValue();
            }
        }

        setExplicitness(new Boolean(isExplicit));
    }

    /**
     * Evaluate
     * @param variableList VariableList object parameter
     * @return Value object
     */
    public Value evaluate(final VariableList variableList) {
        if (!getExplicitness().booleanValue()) {
            throw new IllegalStateException("Attribute: Explicitness is false");
        }

        if (!getCompleteness().booleanValue()) {
            throw new IllegalStateException("Attribute: Completeness is false");
        }

        for (int i = 0; i < attributes.size(); i++) {
            Attribute attribute = (Attribute) attributes.get(i);
            attribute.evaluate(variableList);
        }

        if (attributes.size() > 0) {
            calculateValue();
        } else {
            Variable variable = variableList.findVariable(getName());

            if (variable == null) {
                throw new IllegalArgumentException(
                    "Attribute: Missing variable");
            }

            Value scaleValue = scale.findValue(variable.getValue());
            setValue(scaleValue);
        }

        return getValue();
    }

    /**
     * Get completeness
     * @return Boolean object
     */
    public Boolean getCompleteness() {
        return this.completeness;
    }

    /**
     * Get description
     * @return String object
     */
    public String getDescription() {
        return this.description;
    }

    /**
     * Get explicitness
     * @return Boolean object
     */
    public Boolean getExplicitness() {
        return this.explicitness;
    }

    /**
     * Get name
     * @return String object
     */
    public String getName() {
        return this.name;
    }

    /**
     * Get size
     * @return Integer object
     */
    public Integer getSize() {
        if (scale == null) {
            return new Integer(0);
        } else {
            return this.scale.getSize();
        }
    }

    /**
     * Get value
     * @return Value object
     */
    public Value getValue() {
        return this.value;
    }

    /**
     * Calculate value
     *
     */
    protected void calculateValue() {
        int index = 0;
        int factor = 1;

        if (attributes.size() > 0) {
            for (int i = attributes.size() - 1; i >= 0; i--) {
                int subOrdinal = ((Attribute) attributes.get(i)).getValue()
                                  .getOrdinal().intValue();
                int subSize = ((Attribute) attributes.get(i)).getSize()
                               .intValue();
                index = index + (factor * subOrdinal);
                factor = factor * subSize;
            }

            Rule rule = (Rule) function.get(index);
            setValue(scale.findValue(rule.getLow()));
        }
    }

    /**
     * Set description value
     * @param aDescription String object parameter
     */
    protected void setDescription(final String aDescription) {
        this.description = aDescription;
    }

    /**
     * Set explicitness value
     * @param aExplicitness Boolean object parameter
     */
    protected void setExplicitness(final Boolean aExplicitness) {
        this.explicitness = aExplicitness;
    }

    /**
     * Set name value
     * @param aName String object parameter
     */
    protected void setName(final String aName) {
        this.name = aName;
    }

    /**
     * Set value
     * @param aValue Value object parameter
     */
    protected void setValue(final Value aValue) {
        this.value = aValue;
    }

    /**
     * Set completeness value
     * @param aCompleteness Boolean object parameter
     */
    protected void setCompleteness(final Boolean aCompleteness) {
        this.completeness = aCompleteness;
    }

    /**
     * Print object values to OutputStream
     * @param stream OutputStream object parameter
     * @throws IOException Throws exception if stream cannot write
     */
    public void print(final OutputStream stream) throws IOException {
        StringBuffer sb = new StringBuffer();
        sb.append("Attribute: \n");
        sb.append("Name: ");
        sb.append(getName() + "\n");
        sb.append("Description: ");
        sb.append(getDescription() + "\n");
        sb.append("Completeness: ");
        sb.append(getCompleteness() + "\n");
        sb.append("Explicitness: ");
        sb.append(getExplicitness() + "\n");
        sb.append("Value: ");
        stream.write(sb.toString().getBytes());

        if (value != null) {
            value.print(stream);
        }

        for (int i = 0; i < attributes.size(); i++) {
            stream.write(("SubAttributes[" + i + "]: \n").getBytes());

            Attribute attribute = (Attribute) attributes.get(i);
            attribute.print(stream);
        }

        if (function != null) {
            for (int f = 0; f < function.size(); f++) {
                stream.write(("Rules[" + f + "] \n").getBytes());

                Rule rule = (Rule) function.get(f);
                rule.print(stream);
            }
        }

        scale.print(stream);
    }
}
