////////////////////////////////////////////////////////////////////////////////
//DEXi: Implements a Multi-Attribute Decision Making algorithm developed by
//dr. Marko Bohanec (http://www-ai.ijs.si/MarkoBohanec/dex.html)
//
//Copyright (C) 2004 Asobi d.o.o. (www.asobi.si)
//Authors: dr. Marko Bohanec, Dusan Omercevic, Andrej Kogovsek
//
//This library is free software; you can redistribute it and/or
//modify it under the terms of the GNU Lesser General Public
//License as published by the Free Software Foundation; either
//version 2.1 of the License, or (at your option) any later version.
//
//This library is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
//Lesser General Public License for more details.
//
//You should have received a copy of the GNU Lesser General Public
//License along with this library; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
////////////////////////////////////////////////////////////////////////////////
package si.DEXi;

import org.w3c.dom.*;

import org.xml.sax.InputSource;

import java.io.IOException;
import java.io.OutputStream;
import java.io.StringReader;

import java.util.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;


/**
 * Model class
 * @author Andrej Kogov\u0161ek, Du\u0161an Omer\u010Devi\u0107
 * @version 1.0
 */
public class Model {
    private Vector attributes;

    /**
     * Constructor
     * @param xml String object parameter
     */
    public Model(final String xml) {
        parseXML(xml);
    }

    /**
     * Evaluate method
     * @param strAttribute String object parameter
     * @param variables String object parameter
     * @return value
     */
    public Value evaluate(final String strAttribute, final String variables) {
        if (strAttribute == null) {
            throw new IllegalArgumentException("Model: Missing attribute name");
        }

        VariableList variableList = new VariableList(variables);
        Attribute attribute = findAttribute(strAttribute);

        if (attribute == null) {
            throw new IllegalArgumentException("Model: Unknown attribute name");
        }

        Value value = attribute.evaluate(variableList);

        return value;
    }

    /**
     * Find attribute by name
     * @param name String object parameter
     * @return Attribute
     */
    protected Attribute findAttribute(final String name) {
        Attribute attribute = null;

        for (int i = 0; i < attributes.size(); i++) {
            Attribute attr = (Attribute) attributes.get(i);

            if (attr.getName().equals(name)) {
                attribute = attr;
            }
        }

        return attribute;
    }

    /**
     * Get explicitness
     * @return Boolean object
     */
    public Boolean getExplicitness() {
        boolean isExplicit = true;

        for (int i = 0; i < attributes.size(); i++) {
            Attribute attribute = (Attribute) attributes.get(i);
            isExplicit = isExplicit
                && attribute.getExplicitness().booleanValue();
        }

        return new Boolean(isExplicit);
    }

    /**
     * Get completeness
     * @return Boolean object
     */
    public Boolean getCompleteness() {
        boolean isComplet = true;

        for (int i = 0; i < attributes.size(); i++) {
            Attribute attribute = (Attribute) attributes.get(i);
            isComplet = isComplet && attribute.getCompleteness().booleanValue();
        }

        return new Boolean(isComplet);
    }

    /**
     * Parse XML
     * @param xml String object parameter
     */
    protected void parseXML(final String xml) {
        Element element = null;

        try {
            DocumentBuilderFactory domFactory = DocumentBuilderFactory
                .newInstance();
            DocumentBuilder domBuilder = domFactory.newDocumentBuilder();
            element = domBuilder.parse(new InputSource(new StringReader(xml)))
                                .getDocumentElement();
        } catch (Exception e) {
            throw new IllegalArgumentException("Model: Invalid xml data");
        }

        if (!element.getNodeName().equals("DEXi")) {
            throw new IllegalArgumentException("Model: Xml root must be DEXi");
        }

        NodeList nodes = element.getChildNodes();
        attributes = new Vector();

        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);

            if (node.getNodeName().equals("ATTRIBUTE")) {
                Attribute attribute = new Attribute((Element) node);
                attributes.add(attribute);
            }
        }
    }

    /**
     * Print object values to OutputStream
     * @param stream OutputStream object parameter
     * @throws IOException Throws exception if stream cannot write
     */
    public void print(final OutputStream stream) throws IOException {
        stream.write("Model: \n".getBytes());

        for (int i = 0; i < attributes.size(); i++) {
            stream.write(("Attribute[" + i + "]: \n").getBytes());

            Attribute attribute = (Attribute) attributes.get(i);
            attribute.print(stream);
        }
    }
}
